"""Query conditions and orders for Fair Share repository."""

from __future__ import annotations

import uuid
from collections.abc import Collection

import sqlalchemy as sa

from ai.backend.manager.data.group.types import ProjectType
from ai.backend.manager.data.user.types import UserStatus
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.models.fair_share import (
    DomainFairShareRow,
    ProjectFairShareRow,
    UserFairShareRow,
)
from ai.backend.manager.models.group import GroupRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base.types import QueryCondition, QueryOrder


class DomainFairShareConditions:
    """Query conditions for DomainFairShareRow."""

    @staticmethod
    def by_resource_group(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_domain_name(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_resource_group_contains(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.resource_group.like(f"%{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_equals(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_resource_group_starts_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.resource_group.like(f"{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_ends_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.resource_group.like(f"%{resource_group}")

        return inner

    @staticmethod
    def by_domain_name_contains(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.domain_name.like(f"%{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_equals(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_domain_name_starts_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.domain_name.like(f"{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_ends_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainFairShareRow.domain_name.like(f"%{domain_name}")

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DomainFairShareRow.created_at)
                .where(DomainFairShareRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return DomainFairShareRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DomainFairShareRow.created_at)
                .where(DomainFairShareRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return DomainFairShareRow.created_at > subquery

        return inner

    @staticmethod
    def by_domain_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainRow.is_active == is_active

        return inner


class DomainFairShareOrders:
    """Query orders for DomainFairShareRow."""

    @staticmethod
    def by_fair_share_factor(ascending: bool = False) -> QueryOrder:
        col = DomainFairShareRow.fair_share_factor
        return col.asc() if ascending else col.desc()

    @staticmethod
    def by_domain_name(ascending: bool = True) -> QueryOrder:
        col = DomainFairShareRow.domain_name
        return col.asc() if ascending else col.desc()

    @staticmethod
    def by_created_at(ascending: bool = True) -> QueryOrder:
        col = DomainFairShareRow.created_at
        return col.asc() if ascending else col.desc()

    @staticmethod
    def by_domain_is_active(ascending: bool = True) -> QueryOrder:
        col = DomainRow.is_active
        return col.asc() if ascending else col.desc()


class ProjectFairShareConditions:
    """Query conditions for ProjectFairShareRow."""

    @staticmethod
    def by_resource_group(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_project_id(project_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.project_id == project_id

        return inner

    @staticmethod
    def by_project_ids(project_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.project_id.in_(project_ids)

        return inner

    @staticmethod
    def by_domain_name(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_resource_group_contains(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.resource_group.like(f"%{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_equals(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_resource_group_starts_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.resource_group.like(f"{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_ends_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.resource_group.like(f"%{resource_group}")

        return inner

    @staticmethod
    def by_domain_name_contains(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.domain_name.like(f"%{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_equals(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_domain_name_starts_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.domain_name.like(f"{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_ends_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectFairShareRow.domain_name.like(f"%{domain_name}")

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(ProjectFairShareRow.created_at)
                .where(ProjectFairShareRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return ProjectFairShareRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(ProjectFairShareRow.created_at)
                .where(ProjectFairShareRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return ProjectFairShareRow.created_at > subquery

        return inner

    @staticmethod
    def by_project_name_contains(name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return GroupRow.name.like(f"%{name}%")

        return inner

    @staticmethod
    def by_project_name_equals(name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return GroupRow.name == name

        return inner

    @staticmethod
    def by_project_name_starts_with(name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return GroupRow.name.like(f"{name}%")

        return inner

    @staticmethod
    def by_project_name_ends_with(name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return GroupRow.name.like(f"%{name}")

        return inner

    @staticmethod
    def by_project_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return GroupRow.is_active == is_active

        return inner

    @staticmethod
    def by_project_type_equals(project_type: ProjectType) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return GroupRow.type == project_type

        return inner

    @staticmethod
    def by_project_type_in(project_types: Collection[ProjectType]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return GroupRow.type.in_(project_types)

        return inner


class ProjectFairShareOrders:
    """Query orders for ProjectFairShareRow."""

    @staticmethod
    def by_fair_share_factor(ascending: bool = False) -> QueryOrder:
        col = ProjectFairShareRow.fair_share_factor
        return col.asc() if ascending else col.desc()

    @staticmethod
    def by_created_at(ascending: bool = True) -> QueryOrder:
        col = ProjectFairShareRow.created_at
        return col.asc() if ascending else col.desc()

    @staticmethod
    def by_project_name(ascending: bool = True) -> QueryOrder:
        col = GroupRow.name
        return col.asc() if ascending else col.desc()

    @staticmethod
    def by_project_is_active(ascending: bool = True) -> QueryOrder:
        col = GroupRow.is_active
        return col.asc() if ascending else col.desc()


class UserFairShareConditions:
    """Query conditions for UserFairShareRow."""

    @staticmethod
    def by_resource_group(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_user_uuid(user_uuid: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.user_uuid == user_uuid

        return inner

    @staticmethod
    def by_project_id(project_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.project_id == project_id

        return inner

    @staticmethod
    def by_user_uuids(user_uuids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.user_uuid.in_(user_uuids)

        return inner

    @staticmethod
    def by_project_ids(project_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.project_id.in_(project_ids)

        return inner

    @staticmethod
    def by_domain_name(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_resource_group_contains(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.resource_group.like(f"%{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_equals(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_resource_group_starts_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.resource_group.like(f"{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_ends_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.resource_group.like(f"%{resource_group}")

        return inner

    @staticmethod
    def by_domain_name_contains(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.domain_name.like(f"%{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_equals(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_domain_name_starts_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.domain_name.like(f"{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_ends_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserFairShareRow.domain_name.like(f"%{domain_name}")

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(UserFairShareRow.created_at)
                .where(UserFairShareRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return UserFairShareRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(UserFairShareRow.created_at)
                .where(UserFairShareRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return UserFairShareRow.created_at > subquery

        return inner

    @staticmethod
    def by_user_username_contains(username: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.username.like(f"%{username}%")

        return inner

    @staticmethod
    def by_user_username_equals(username: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.username == username

        return inner

    @staticmethod
    def by_user_username_starts_with(username: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.username.like(f"{username}%")

        return inner

    @staticmethod
    def by_user_username_ends_with(username: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.username.like(f"%{username}")

        return inner

    @staticmethod
    def by_user_email_contains(email: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.email.like(f"%{email}%")

        return inner

    @staticmethod
    def by_user_email_equals(email: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.email == email

        return inner

    @staticmethod
    def by_user_email_starts_with(email: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.email.like(f"{email}%")

        return inner

    @staticmethod
    def by_user_email_ends_with(email: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.email.like(f"%{email}")

        return inner

    @staticmethod
    def by_user_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if is_active:
                return UserRow.status == UserStatus.ACTIVE
            return UserRow.status != UserStatus.ACTIVE

        return inner


class UserFairShareOrders:
    """Query orders for UserFairShareRow."""

    @staticmethod
    def by_fair_share_factor(ascending: bool = False) -> QueryOrder:
        col = UserFairShareRow.fair_share_factor
        return col.asc() if ascending else col.desc()

    @staticmethod
    def by_created_at(ascending: bool = True) -> QueryOrder:
        col = UserFairShareRow.created_at
        return col.asc() if ascending else col.desc()

    @staticmethod
    def by_user_username(ascending: bool = True) -> QueryOrder:
        col = UserRow.username
        return col.asc() if ascending else col.desc()

    @staticmethod
    def by_user_email(ascending: bool = True) -> QueryOrder:
        col = UserRow.email
        return col.asc() if ascending else col.desc()



from __future__ import annotations

import abc
import logging
from dataclasses import dataclass
from typing import Any, TypedDict, override

import aiohttp
import yarl

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.errors.common import (
    GenericBadRequest,
    InternalServerError,
    ObjectNotFound,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class HarborProjectInfo:
    url: str
    project: str
    ssl_verify: bool


class HarborAuthArgs(TypedDict):
    username: str
    password: str


class HarborProjectQuotaInfo(TypedDict):
    previous_quota: int
    quota_id: int


def _get_harbor_auth_args(auth_args: HarborAuthArgs) -> dict[str, Any]:
    return {"auth": aiohttp.BasicAuth(auth_args["username"], auth_args["password"])}


class AbstractPerProjectRegistryQuotaClient(abc.ABC):
    async def create_quota(
        self, project_info: HarborProjectInfo, quota: int, auth_args: HarborAuthArgs
    ) -> None:
        raise NotImplementedError

    async def update_quota(
        self, project_info: HarborProjectInfo, quota: int, auth_args: HarborAuthArgs
    ) -> None:
        raise NotImplementedError

    async def delete_quota(
        self, project_info: HarborProjectInfo, auth_args: HarborAuthArgs
    ) -> None:
        raise NotImplementedError

    async def read_quota(self, project_info: HarborProjectInfo, auth_args: HarborAuthArgs) -> int:
        raise NotImplementedError


class PerProjectHarborQuotaClient(AbstractPerProjectRegistryQuotaClient):
    async def _get_harbor_project_id(
        self,
        sess: aiohttp.ClientSession,
        project_info: HarborProjectInfo,
        rqst_args: dict[str, Any],
    ) -> str:
        get_project_id_api = (
            yarl.URL(project_info.url) / "api" / "v2.0" / "projects" / project_info.project
        )

        async with sess.get(get_project_id_api, allow_redirects=False, **rqst_args) as resp:
            if resp.status != 200:
                raise InternalServerError(f"Failed to get harbor project_id! response: {resp}")

            res = await resp.json()
            harbor_project_id = res["project_id"]
            return str(harbor_project_id)

    async def _get_quota_info(
        self,
        sess: aiohttp.ClientSession,
        project_info: HarborProjectInfo,
        rqst_args: dict[str, Any],
    ) -> HarborProjectQuotaInfo:
        harbor_project_id = await self._get_harbor_project_id(sess, project_info, rqst_args)
        get_quota_id_api = (yarl.URL(project_info.url) / "api" / "v2.0" / "quotas").with_query({
            "reference": "project",
            "reference_id": harbor_project_id,
        })

        async with sess.get(get_quota_id_api, allow_redirects=False, **rqst_args) as resp:
            if resp.status != 200:
                raise InternalServerError(f"Failed to get quota info! response: {resp}")

            res = await resp.json()
            if not res:
                raise ObjectNotFound(object_name="quota entity")
            if len(res) > 1:
                raise InternalServerError(
                    f"Multiple quota entities found. (project_id: {harbor_project_id})"
                )

            previous_quota = res[0]["hard"]["storage"]
            quota_id = res[0]["id"]
            return HarborProjectQuotaInfo(previous_quota=previous_quota, quota_id=quota_id)

    @override
    async def read_quota(self, project_info: HarborProjectInfo, auth_args: HarborAuthArgs) -> int:
        connector = aiohttp.TCPConnector(ssl=project_info.ssl_verify)
        async with aiohttp.ClientSession(connector=connector) as sess:
            rqst_args = _get_harbor_auth_args(auth_args)
            quota_info = await self._get_quota_info(sess, project_info, rqst_args)
            previous_quota = quota_info["previous_quota"]
            if previous_quota == -1:
                raise ObjectNotFound(object_name="quota entity")
            return previous_quota

    @override
    async def create_quota(
        self, project_info: HarborProjectInfo, quota: int, auth_args: HarborAuthArgs
    ) -> None:
        connector = aiohttp.TCPConnector(ssl=project_info.ssl_verify)
        async with aiohttp.ClientSession(connector=connector) as sess:
            rqst_args = _get_harbor_auth_args(auth_args)
            quota_info = await self._get_quota_info(sess, project_info, rqst_args)
            previous_quota, quota_id = quota_info["previous_quota"], quota_info["quota_id"]

            if previous_quota != -1:
                raise GenericBadRequest("Quota limit already exists!")

            put_quota_api = yarl.URL(project_info.url) / "api" / "v2.0" / "quotas" / str(quota_id)
            payload = {"hard": {"storage": quota}}

            async with sess.put(
                put_quota_api, json=payload, allow_redirects=False, **rqst_args
            ) as resp:
                if resp.status != 200:
                    log.error(f"Failed to create quota! response: {resp}")
                    raise InternalServerError(f"Failed to create quota! response: {resp}")

    @override
    async def update_quota(
        self, project_info: HarborProjectInfo, quota: int, auth_args: HarborAuthArgs
    ) -> None:
        connector = aiohttp.TCPConnector(ssl=project_info.ssl_verify)
        async with aiohttp.ClientSession(connector=connector) as sess:
            rqst_args = _get_harbor_auth_args(auth_args)
            quota_info = await self._get_quota_info(sess, project_info, rqst_args)
            previous_quota, quota_id = quota_info["previous_quota"], quota_info["quota_id"]

            if previous_quota == -1:
                raise ObjectNotFound(object_name="quota entity")

            put_quota_api = yarl.URL(project_info.url) / "api" / "v2.0" / "quotas" / str(quota_id)
            payload = {"hard": {"storage": quota}}

            async with sess.put(
                put_quota_api, json=payload, allow_redirects=False, **rqst_args
            ) as resp:
                if resp.status != 200:
                    log.error(f"Failed to update quota! response: {resp}")
                    raise InternalServerError(f"Failed to update quota! response: {resp}")

    @override
    async def delete_quota(
        self, project_info: HarborProjectInfo, auth_args: HarborAuthArgs
    ) -> None:
        connector = aiohttp.TCPConnector(ssl=project_info.ssl_verify)
        async with aiohttp.ClientSession(connector=connector) as sess:
            rqst_args = _get_harbor_auth_args(auth_args)
            quota_info = await self._get_quota_info(sess, project_info, rqst_args)
            previous_quota, quota_id = quota_info["previous_quota"], quota_info["quota_id"]

            if previous_quota == -1:
                raise ObjectNotFound(object_name="quota entity")

            put_quota_api = yarl.URL(project_info.url) / "api" / "v2.0" / "quotas" / str(quota_id)
            payload = {"hard": {"storage": -1}}

            async with sess.put(
                put_quota_api, json=payload, allow_redirects=False, **rqst_args
            ) as resp:
                if resp.status != 200:
                    log.error(f"Failed to delete quota! response: {resp}")
                    raise InternalServerError(f"Failed to delete quota! response: {resp}")



"""Add wsproxy_addr column on scaling_group

Revision ID: 60a1effa77d2
Revises: 8679d0a7e22b
Create Date: 2021-09-17 13:19:57.525513

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "60a1effa77d2"
down_revision = "8679d0a7e22b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "scaling_groups", sa.Column("wsproxy_addr", sa.String(length=1024), nullable=True)
    )


def downgrade() -> None:
    op.drop_column("scaling_groups", "wsproxy_addr")



"""Data classes for Fair Share domain."""

from __future__ import annotations

import uuid
from collections.abc import Mapping
from dataclasses import dataclass, field
from datetime import date, datetime, timedelta
from decimal import Decimal

from ai.backend.common.types import ResourceSlot


@dataclass(frozen=True)
class ProjectUserIds:
    """A project with its associated user IDs for batch fair share lookup."""

    project_id: uuid.UUID
    user_ids: frozenset[uuid.UUID]


@dataclass(frozen=True)
class UserProjectKey:
    """Key identifying a user within a project for fair share tracking."""

    user_uuid: uuid.UUID
    project_id: uuid.UUID


@dataclass(frozen=True)
class FairShareSpec:
    """Specification parameters for fair share calculation.

    These values determine how fair share factor is computed.
    """

    weight: Decimal
    """Base weight for this entity (higher weight = higher priority).
    This is always set - either explicitly or from resource group's default_weight."""

    half_life_days: int
    """Half-life for exponential decay in days."""

    lookback_days: int
    """Total lookback period in days for usage aggregation."""

    decay_unit_days: int
    """Granularity of decay buckets in days."""

    resource_weights: ResourceSlot
    """Weights for each resource type when calculating normalized usage."""


@dataclass(frozen=True)
class FairShareCalculationSnapshot:
    """Snapshot of the most recent fair share calculation.

    Contains the computed values and the time window used for calculation.
    """

    fair_share_factor: Decimal
    """Computed fair share factor (0-1 range, higher = more entitled)."""

    total_decayed_usage: ResourceSlot
    """Sum of decayed historical usage across all resource types."""

    normalized_usage: Decimal
    """Single scalar representing weighted resource consumption."""

    lookback_start: date
    """Start date of the lookback window used in calculation."""

    lookback_end: date
    """End date of the lookback window used in calculation."""

    last_calculated_at: datetime
    """Timestamp when this calculation was performed."""

    @classmethod
    def create_default(
        cls,
        lookback_days: int,
        available_slots: ResourceSlot,
        now: datetime,
    ) -> FairShareCalculationSnapshot:
        """Create default snapshot with zero usage.

        Args:
            lookback_days: Number of days to look back for calculation
            available_slots: Scaling group's available slots (for resource keys)
            now: Current datetime for lookback window calculation

        Returns:
            FairShareCalculationSnapshot with zero usage and default values
        """
        today = now.date()
        lookback_start = today - timedelta(days=lookback_days)

        # Initialize total_decayed_usage with zeros based on available_slots keys
        zero_usage = ResourceSlot({key: Decimal("0.0") for key in available_slots})

        return cls(
            fair_share_factor=Decimal("1.0"),  # No usage history
            total_decayed_usage=zero_usage,  # Zero values with correct keys
            normalized_usage=Decimal("0.0"),
            lookback_start=lookback_start,
            lookback_end=today,
            last_calculated_at=now,
        )


@dataclass(frozen=True)
class FairShareMetadata:
    """Audit metadata for fair share entities."""

    created_at: datetime
    updated_at: datetime


@dataclass(frozen=True)
class FairShareData:
    """Fair share calculation data (reusable across Domain/Project/User)."""

    spec: FairShareSpec
    """Specification parameters for fair share calculation."""

    calculation_snapshot: FairShareCalculationSnapshot
    """Snapshot of the most recent fair share calculation."""

    metadata: FairShareMetadata | None
    """Audit metadata. None for default-generated records."""

    use_default: bool
    """True if this record uses default weight (generated from scaling group defaults)."""

    uses_default_resources: frozenset[str] = frozenset()
    """Resource types that use default weight from scaling group.
    Empty frozenset means all resources have explicit weights configured."""


@dataclass(frozen=True)
class DomainFairShareData:
    """Domain-level fair share data."""

    resource_group: str
    domain_name: str
    data: FairShareData
    """Nested fair share data structure."""


@dataclass(frozen=True)
class ProjectFairShareData:
    """Project-level fair share data."""

    resource_group: str
    project_id: uuid.UUID
    domain_name: str
    data: FairShareData
    """Nested fair share data structure."""


@dataclass(frozen=True)
class UserFairShareData:
    """User-level fair share data."""

    resource_group: str
    user_uuid: uuid.UUID
    project_id: uuid.UUID
    domain_name: str
    data: FairShareData
    """Nested fair share data structure."""
    scheduling_rank: int | None = None
    """Computed scheduling priority rank. Lower value = higher priority.
    None means rank calculation has not been performed yet."""


@dataclass(frozen=True)
class UserFairShareFactors:
    """Combined fair share factors for a user across all hierarchy levels.

    Used for sorting workloads by fair share priority.
    Higher factors = higher priority.
    """

    user_uuid: uuid.UUID
    project_id: uuid.UUID
    domain_name: str
    domain_factor: Decimal
    project_factor: Decimal
    user_factor: Decimal

    def sort_key(self) -> tuple[Decimal, Decimal, Decimal]:
        """Return a tuple for sorting (higher factor = higher priority).

        Returns negated values since Python sorts ascending by default.
        """
        return (-self.domain_factor, -self.project_factor, -self.user_factor)


@dataclass(frozen=True)
class DomainFairShareSearchResult:
    """Search result with pagination info for domain fair shares."""

    items: list[DomainFairShareData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass(frozen=True)
class ProjectFairShareSearchResult:
    """Search result with pagination info for project fair shares."""

    items: list[ProjectFairShareData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass(frozen=True)
class UserFairShareSearchResult:
    """Search result with pagination info for user fair shares."""

    items: list[UserFairShareData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


# ==================== Batched Read Results ====================


@dataclass(frozen=True)
class FairSharesByLevel:
    """Fair share records grouped by hierarchy level.

    Used for factor calculation to get current weights and configurations.
    """

    domain: Mapping[str, DomainFairShareData]
    """Domain fair shares keyed by domain_name."""

    project: Mapping[uuid.UUID, ProjectFairShareData]
    """Project fair shares keyed by project_id."""

    user: Mapping[UserProjectKey, UserFairShareData]
    """User fair shares keyed by UserProjectKey."""


@dataclass(frozen=True)
class RawUsageBucketsByLevel:
    """Raw usage buckets grouped by hierarchy level before decay is applied.

    Contains per-date ResourceSlot buckets for each entity in the lookback window.
    The Calculator is responsible for applying time decay to these raw buckets.
    """

    domain: Mapping[str, Mapping[date, ResourceSlot]]
    """Raw usage buckets by domain_name -> date -> usage."""

    project: Mapping[uuid.UUID, Mapping[date, ResourceSlot]]
    """Raw usage buckets by project_id -> date -> usage."""

    user: Mapping[UserProjectKey, Mapping[date, ResourceSlot]]
    """Raw usage buckets by UserProjectKey -> date -> usage."""

    def is_empty(self) -> bool:
        """Check if there are no usage buckets at any level."""
        return not self.domain and not self.project and not self.user


@dataclass(frozen=True)
class FairShareCalculationContext:
    """All data needed for fair share factor calculation.

    Combines fair share records (for weights) and raw usage buckets.
    This is the result of batched DB reads.
    The Calculator applies time decay to raw_usage_buckets internally.
    """

    fair_shares: FairSharesByLevel
    """Current fair share records with weights."""

    raw_usage_buckets: RawUsageBucketsByLevel
    """Raw usage buckets from lookback window (decay not applied yet)."""

    half_life_days: int
    """Half-life for exponential decay in days."""

    lookback_days: int
    """Number of days in lookback window."""

    default_weight: Decimal
    """Default weight for entities without explicit weight."""

    resource_weights: ResourceSlot
    """Default resource weights for normalized usage calculation."""

    cluster_capacity: ResourceSlot
    """Total available slots from all ALIVE schedulable agents in the scaling group.
    Used to normalize usage: usage[r] / (capacity[r] * lookback_days * SECONDS_PER_DAY).
    """

    today: date
    """Current date for decay calculation."""

    project_domain_names: Mapping[uuid.UUID, str] = field(default_factory=dict)
    """Mapping from project_id to domain_name.
    Used as fallback when fair_shares.project doesn't contain the project
    (new projects with usage but no fair share record yet)."""



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.clients.valkey_client.valkey_artifact.client import (
    ValkeyArtifactDownloadTrackingClient,
)
from ai.backend.common.clients.valkey_client.valkey_bgtask.client import ValkeyBgtaskClient
from ai.backend.common.clients.valkey_client.valkey_container_log.client import (
    ValkeyContainerLogClient,
)
from ai.backend.common.clients.valkey_client.valkey_image.client import ValkeyImageClient
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_schedule.client import ValkeyScheduleClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.clients.valkey_client.valkey_stream.client import ValkeyStreamClient
from ai.backend.common.defs import (
    REDIS_BGTASK_DB,
    REDIS_CONTAINER_LOG,
    REDIS_IMAGE_DB,
    REDIS_LIVE_DB,
    REDIS_STATISTICS_DB,
    REDIS_STREAM_DB,
    RedisRole,
)
from ai.backend.common.health_checker import (
    CID_REDIS_ARTIFACT,
    CID_REDIS_BGTASK,
    CID_REDIS_CONTAINER_LOG,
    CID_REDIS_IMAGE,
    CID_REDIS_LIVE,
    CID_REDIS_SCHEDULE,
    CID_REDIS_STAT,
    CID_REDIS_STREAM,
    ServiceHealthChecker,
)
from ai.backend.common.health_checker.checkers.valkey import ValkeyHealthChecker
from ai.backend.manager.config.unified import ManagerUnifiedConfig

from .base import InfrastructureDependency


@dataclass
class ValkeyClients:
    """Container for all specialized Valkey client instances."""

    artifact: ValkeyArtifactDownloadTrackingClient
    container_log: ValkeyContainerLogClient
    live: ValkeyLiveClient
    stat: ValkeyStatClient
    image: ValkeyImageClient
    stream: ValkeyStreamClient
    schedule: ValkeyScheduleClient
    bgtask: ValkeyBgtaskClient

    async def close(self) -> None:
        """Close all Valkey client connections."""
        await self.artifact.close()
        await self.container_log.close()
        await self.image.close()
        await self.stat.close()
        await self.live.close()
        await self.stream.close()
        await self.schedule.close()
        await self.bgtask.close()


class ValkeyDependency(InfrastructureDependency[ValkeyClients]):
    """Provides lifecycle management for 8 specialized Valkey clients."""

    @property
    def stage_name(self) -> str:
        return "valkey"

    @asynccontextmanager
    async def provide(self, setup_input: ManagerUnifiedConfig) -> AsyncIterator[ValkeyClients]:
        """Initialize and provide all Valkey clients.

        Args:
            setup_input: Configuration containing redis settings

        Yields:
            ValkeyClients instance containing all 8 specialized clients
        """
        valkey_profile_target = setup_input.redis.to_valkey_profile_target()

        # Create all 8 specialized clients
        clients = ValkeyClients(
            artifact=await ValkeyArtifactDownloadTrackingClient.create(
                valkey_profile_target.profile_target(RedisRole.STATISTICS),
                db_id=REDIS_STATISTICS_DB,
                human_readable_name="artifact",
            ),
            container_log=await ValkeyContainerLogClient.create(
                valkey_profile_target.profile_target(RedisRole.CONTAINER_LOG),
                db_id=REDIS_CONTAINER_LOG,
                human_readable_name="container_log",
            ),
            live=await ValkeyLiveClient.create(
                valkey_profile_target.profile_target(RedisRole.LIVE),
                db_id=REDIS_LIVE_DB,
                human_readable_name="live",
            ),
            stat=await ValkeyStatClient.create(
                valkey_profile_target.profile_target(RedisRole.STATISTICS),
                db_id=REDIS_STATISTICS_DB,
                human_readable_name="stat",
            ),
            image=await ValkeyImageClient.create(
                valkey_profile_target.profile_target(RedisRole.IMAGE),
                db_id=REDIS_IMAGE_DB,
                human_readable_name="image",
            ),
            stream=await ValkeyStreamClient.create(
                valkey_profile_target.profile_target(RedisRole.STREAM),
                human_readable_name="stream",
                db_id=REDIS_STREAM_DB,
            ),
            schedule=await ValkeyScheduleClient.create(
                valkey_profile_target.profile_target(RedisRole.STREAM),
                db_id=REDIS_LIVE_DB,
                human_readable_name="schedule",
            ),
            bgtask=await ValkeyBgtaskClient.create(
                valkey_profile_target.profile_target(RedisRole.BGTASK),
                human_readable_name="bgtask",
                db_id=REDIS_BGTASK_DB,
            ),
        )

        # Health check
        await clients.live.get_server_time()

        try:
            yield clients
        finally:
            await clients.close()

    def gen_health_checkers(
        self,
        resource: ValkeyClients,
    ) -> ServiceHealthChecker:
        """
        Return a single health checker for all 8 Valkey clients.

        Args:
            resource: The initialized Valkey clients

        Returns:
            ValkeyHealthChecker that checks all 8 Valkey clients
        """
        return ValkeyHealthChecker(
            clients={
                CID_REDIS_ARTIFACT: resource.artifact,
                CID_REDIS_CONTAINER_LOG: resource.container_log,
                CID_REDIS_LIVE: resource.live,
                CID_REDIS_STAT: resource.stat,
                CID_REDIS_IMAGE: resource.image,
                CID_REDIS_STREAM: resource.stream,
                CID_REDIS_SCHEDULE: resource.schedule,
                CID_REDIS_BGTASK: resource.bgtask,
            }
        )



"""add-index-for-cluster_role

Revision ID: 518ecf41f567
Revises: dc9b66466e43
Create Date: 2021-01-07 00:04:53.794638

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "518ecf41f567"
down_revision = "dc9b66466e43"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(op.f("ix_kernels_cluster_role"), "kernels", ["cluster_role"], unique=False)
    op.create_index("ix_kernels_status_role", "kernels", ["status", "cluster_role"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_kernels_status_role", table_name="kernels")
    op.drop_index(op.f("ix_kernels_cluster_role"), table_name="kernels")
    # ### end Alembic commands ###



"""add group model

Revision ID: 7a82e0c70122
Revises: bae1a7326e8a
Create Date: 2019-05-09 10:00:55.788734

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "7a82e0c70122"
down_revision = "bae1a7326e8a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "groups",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.Column("description", sa.String(length=512), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column(
            "modified_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("domain_name", sa.String(length=64), nullable=False),
        sa.ForeignKeyConstraint(
            ["domain_name"],
            ["domains.name"],
            name=op.f("fk_groups_domain_name_domains"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_groups")),
        sa.UniqueConstraint("name", "domain_name", name="uq_groups_name_domain_name"),
    )
    op.create_index(op.f("ix_groups_domain_name"), "groups", ["domain_name"], unique=False)
    op.create_table(
        "association_groups_users",
        sa.Column("user_id", GUID(), nullable=False),
        sa.Column("group_id", GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["group_id"],
            ["groups.id"],
            name=op.f("fk_association_groups_users_group_id_groups"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["users.uuid"],
            name=op.f("fk_association_groups_users_user_id_users"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.UniqueConstraint("user_id", "group_id", name="uq_association_user_id_group_id"),
    )


def downgrade() -> None:
    op.drop_table("association_groups_users")
    op.drop_index(op.f("ix_groups_domain_name"), table_name="groups")
    op.drop_table("groups")



from dataclasses import dataclass
from typing import override

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.repositories.image.updaters import ImageUpdaterSpec
from ai.backend.manager.services.image.actions.base import ImageAction


@dataclass
class ModifyImageAction(ImageAction):
    target: str
    architecture: str
    updater_spec: ImageUpdaterSpec

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class ModifyImageActionResult(BaseActionResult):
    image: ImageData

    @override
    def entity_id(self) -> str | None:
        return str(self.image.id)


class ModifyImageActionUnknownImageReferenceError(BackendAIError):
    error_type = "https://api.backend.ai/probs/image-not-found"
    error_title = "Unknown image reference."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.IMAGE,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.NOT_FOUND,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import ScopeType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.permission_contoller.actions.base import RoleAction


@dataclass
class GetScopeTypesAction(RoleAction):
    """Action to get available scope types.

    This action is only available to superadmins.
    Permission check is performed at the API handler level.
    """

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class GetScopeTypesActionResult(BaseActionResult):
    """Result of getting scope types."""

    scope_types: list[ScopeType]

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

import datetime as dt
import logging
import uuid
from collections.abc import Iterable, MutableMapping
from datetime import UTC, datetime
from http import HTTPStatus
from typing import TYPE_CHECKING, Any

import aiohttp_cors
import attrs
import sqlalchemy as sa
import trafaret as t
from aiohttp import web
from dateutil.relativedelta import relativedelta

from ai.backend.common import validators as tx
from ai.backend.common.distributed import GlobalTimer
from ai.backend.common.events.dispatcher import EventHandler
from ai.backend.common.events.event_types.log.anycast import DoLogCleanupEvent
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter, LogLevel
from ai.backend.manager.data.error_log.types import ErrorLogSeverity
from ai.backend.manager.defs import LockID
from ai.backend.manager.errors.resource import DBOperationFailed
from ai.backend.manager.models.error_logs import error_logs
from ai.backend.manager.models.group import association_groups_users as agus
from ai.backend.manager.models.group import groups
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.repositories.base import Creator
from ai.backend.manager.repositories.error_log.creators import ErrorLogCreatorSpec
from ai.backend.manager.services.error_log.actions import CreateErrorLogAction

from .auth import auth_required
from .manager import READ_ALLOWED, server_status_required
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params, get_access_key_scopes

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict(
        {
            t.Key("severity"): tx.Enum(LogLevel),
            t.Key("source"): t.String,
            t.Key("message"): t.String,
            t.Key("context_lang"): t.String,
            t.Key("context_env"): tx.JSONString,
            t.Key("request_url", default=None): t.Null | t.String,
            t.Key("request_status", default=None): t.Null | t.Int,
            t.Key("traceback", default=None): t.Null | t.String,
        },
    )
)
async def append(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    params["domain"] = request["user"]["domain_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    requester_uuid = request["user"]["uuid"]
    log.info(
        "CREATE (ak:{0}/{1})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
    )

    severity = ErrorLogSeverity(params["severity"].value.lower())
    creator = Creator(
        spec=ErrorLogCreatorSpec(
            severity=severity,
            source=params["source"],
            user=requester_uuid,
            message=params["message"],
            context_lang=params["context_lang"],
            context_env=params["context_env"],
            request_url=params["request_url"],
            request_status=params["request_status"],
            traceback=params["traceback"],
        )
    )
    action = CreateErrorLogAction(creator=creator)
    await root_ctx.processors.error_log.create.wait_for_complete(action)

    return web.json_response({"success": True})


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key("mark_read", default=False): t.ToBool(),
        t.Key("page_size", default=20): t.ToInt(lt=101),
        t.Key("page_no", default=1): t.ToInt(),
    }),
)
async def list_logs(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    resp: MutableMapping[str, Any] = {"logs": []}
    domain_name = request["user"]["domain_name"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]

    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        "LIST (ak:{0}/{1})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
    )
    async with root_ctx.db.begin() as conn:
        is_admin = True
        select_query = (
            sa.select(error_logs)
            .select_from(error_logs)
            .order_by(sa.desc(error_logs.c.created_at))
            .limit(params["page_size"])
        )
        count_query = sa.select(sa.func.count()).select_from(error_logs)
        if params["page_no"] > 1:
            select_query = select_query.offset((params["page_no"] - 1) * params["page_size"])
        if request["is_superadmin"]:
            pass
        elif user_role == UserRole.ADMIN or user_role == "admin":
            j = groups.join(agus, groups.c.id == agus.c.group_id)
            usr_query = (
                sa.select(agus.c.user_id).select_from(j).where(groups.c.domain_name == domain_name)
            )
            result = await conn.execute(usr_query)
            usrs = result.fetchall()
            user_ids = [g.id for g in usrs]
            where = error_logs.c.user.in_(user_ids)
            select_query = select_query.where(where)
            count_query = count_query.where(where)
        else:
            is_admin = False
            where = (error_logs.c.user == user_uuid) & (~error_logs.c.is_cleared)
            select_query = select_query.where(where)
            count_query = count_query.where(where)

        result = await conn.execute(select_query)
        for row in result:
            result_item = {
                "log_id": str(row.id),
                "created_at": datetime.timestamp(row.created_at),
                "severity": row.severity,
                "source": row.source,
                "user": row.user,
                "is_read": row.is_read,
                "message": row.message,
                "context_lang": row.context_lang,
                "context_env": row.context_env,
                "request_url": row.request_url,
                "request_status": row.request_status,
                "traceback": row.traceback,
            }
            if result_item["user"] is not None:
                result_item["user"] = str(result_item["user"])
            if is_admin:
                result_item["is_cleared"] = row.is_cleared
            resp["logs"].append(result_item)
        resp["count"] = await conn.scalar(count_query)
        if params["mark_read"]:
            read_update_query = (
                sa.update(error_logs)
                .values(is_read=True)
                .where(error_logs.c.id.in_([x["log_id"] for x in resp["logs"]]))
            )
            await conn.execute(read_update_query)
        return web.json_response(resp, status=HTTPStatus.OK)


@auth_required
@server_status_required(READ_ALLOWED)
async def mark_cleared(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    domain_name = request["user"]["domain_name"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    log_id = uuid.UUID(request.match_info["log_id"])

    log.info("CLEAR")
    async with root_ctx.db.begin() as conn:
        update_query = sa.update(error_logs).values(is_cleared=True)
        if request["is_superadmin"]:
            update_query = update_query.where(error_logs.c.id == log_id)
        elif user_role == UserRole.ADMIN or user_role == "admin":
            j = groups.join(agus, groups.c.id == agus.c.group_id)
            usr_query = (
                sa.select(agus.c.user_id).select_from(j).where(groups.c.domain_name == domain_name)
            )
            result = await conn.execute(usr_query)
            usrs = result.fetchall()
            user_ids = [g.id for g in usrs]
            update_query = update_query.where(
                (error_logs.c.user.in_(user_ids)) & (error_logs.c.id == log_id),
            )
        else:
            update_query = update_query.where(
                (error_logs.c.user == user_uuid) & (error_logs.c.id == log_id),
            )

        result = await conn.execute(update_query)
        if result.rowcount != 1:
            raise DBOperationFailed(f"Failed to update error log: {log_id}")

        return web.json_response({"success": True}, status=HTTPStatus.OK)


async def log_cleanup_task(app: web.Application, _src: AgentId, _event: DoLogCleanupEvent) -> None:
    root_ctx: RootContext = app["_root.context"]
    raw_lifetime = await root_ctx.etcd.get("config/logs/error/retention")
    if raw_lifetime is None:
        raw_lifetime = "90d"
    lifetime: dt.timedelta | relativedelta
    try:
        lifetime = tx.TimeDuration().check(raw_lifetime)
    except ValueError:
        lifetime = dt.timedelta(days=90)
        log.warning(
            "Failed to parse the error log retention period ({}) read from etcd; "
            "falling back to 90 days",
            raw_lifetime,
        )
    boundary = datetime.now(UTC) - lifetime
    async with root_ctx.db.begin() as conn:
        query = sa.delete(error_logs).where(error_logs.c.created_at < boundary)
        result = await conn.execute(query)
        if result.rowcount > 0:
            log.info("Cleaned up {} log(s) filed before {}", result.rowcount, boundary)


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    log_cleanup_timer: GlobalTimer
    log_cleanup_timer_evh: EventHandler[web.Application, DoLogCleanupEvent]


async def init(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["logs.context"]
    app_ctx.log_cleanup_timer_evh = root_ctx.event_dispatcher.consume(
        DoLogCleanupEvent,
        app,
        log_cleanup_task,
    )
    app_ctx.log_cleanup_timer = GlobalTimer(
        root_ctx.distributed_lock_factory(LockID.LOCKID_LOG_CLEANUP_TIMER, 20.0),
        root_ctx.event_producer,
        lambda: DoLogCleanupEvent(),
        20.0,
        initial_delay=17.0,
        task_name="log_cleanup_task",
    )
    await app_ctx.log_cleanup_timer.join()


async def shutdown(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["logs.context"]
    await app_ctx.log_cleanup_timer.leave()
    root_ctx.event_dispatcher.unconsume(app_ctx.log_cleanup_timer_evh)


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["api_versions"] = (4, 5)
    app["prefix"] = "logs/error"
    app["logs.context"] = PrivateContext()
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("POST", "", append))
    cors.add(app.router.add_route("GET", "", list_logs))
    cors.add(app.router.add_route("POST", r"/{log_id}/clear", mark_cleared))

    return app, []



from .row import (
    MAXIMUM_DOTFILE_SIZE,
    Dotfile,
    KeyPairRow,
    keypairs,
    query_bootstrap_script,
    query_owned_dotfiles,
    verify_dotfile_name,
)
from .row import (
    generate_keypair as generate_keypair,
)
from .row import (
    generate_keypair_data as generate_keypair_data,
)
from .row import (
    generate_ssh_keypair as generate_ssh_keypair,
)
from .row import (
    prepare_new_keypair as prepare_new_keypair,
)
from .row import (
    validate_ssh_keypair as validate_ssh_keypair,
)

__all__ = (
    "MAXIMUM_DOTFILE_SIZE",
    "Dotfile",
    "KeyPairRow",
    "keypairs",
    "query_bootstrap_script",
    "query_owned_dotfiles",
    "verify_dotfile_name",
)



from __future__ import annotations

from contextlib import AbstractAsyncContextManager
from typing import TYPE_CHECKING, Any

from ai.backend.common.lock import AbstractDistributedLock

from .defs import LockID
from .errors.resource import DBOperationFailed

if TYPE_CHECKING:
    from .models.utils import ExtendedAsyncSAEngine


class PgAdvisoryLock(AbstractDistributedLock):
    _lock_ctx: AbstractAsyncContextManager[None] | None

    def __init__(self, db: ExtendedAsyncSAEngine, lock_id: LockID) -> None:
        self.db = db
        self.lock_id = lock_id
        self._lock_ctx = None

    async def __aenter__(self) -> Any:
        self._lock_ctx = self.db.advisory_lock(self.lock_id)
        await self._lock_ctx.__aenter__()

    async def __aexit__(self, *exc_info: Any) -> bool | None:
        if self._lock_ctx is None:
            raise DBOperationFailed("Lock context is not initialized")
        try:
            return await self._lock_ctx.__aexit__(*exc_info)
        finally:
            self._lock_ctx = None



from __future__ import annotations

from collections.abc import Collection
from dataclasses import dataclass
from uuid import UUID

from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.data.permission.types import (
    OperationType,
)
from ai.backend.manager.data.permission.id import (
    ObjectId,
    ScopeId,
)
from ai.backend.manager.models.rbac_models.permission.object_permission import ObjectPermissionRow

# =============================================================================
# Data Classes
# =============================================================================


@dataclass
class RBACGranter:
    """
    Data class for granting object-level permissions to specific role(s).

    Note: Only entity-level granting is supported. Field-level granting is not supported.

    Attributes:
        granted_entity_id: The entity to grant access to (must be entity, not field).
        granted_entity_scope_id: The original scope where the entity belongs.
        target_role_ids: The role ID(s) to grant permissions to.
        operations: The operations to grant on the entity.
    """

    granted_entity_id: ObjectId
    granted_entity_scope_id: ScopeId
    target_role_ids: list[UUID]
    operations: list[OperationType]


# =============================================================================
# Insert Helpers
# =============================================================================


async def _insert_object_permissions(
    db_sess: SASession,
    role_ids: Collection[UUID],
    entity_id: ObjectId,
    operations: Collection[OperationType],
) -> None:
    """
    Insert object permissions for each role.

    Raises IntegrityError on unique constraint violation (duplicate permission).
    """
    if not role_ids or not operations:
        return

    obj_perms = [
        ObjectPermissionRow(
            role_id=role_id,
            entity_type=entity_id.entity_type,
            entity_id=entity_id.entity_id,
            operation=operation,
        )
        for role_id in role_ids
        for operation in operations
    ]
    db_sess.add_all(obj_perms)
    await db_sess.flush()


# =============================================================================
# Public API
# =============================================================================


async def execute_rbac_granter(
    db_sess: SASession,
    granter: RBACGranter,
) -> None:
    """
    Grant object-level permissions to specified roles.

    This is used when sharing an existing entity with specific roles.
    For example, when user A invites user B to a VFolder:
    - User B's role (provided by caller) gets object permissions for that VFolder.

    Raises:
        IntegrityError: If duplicate object permission already exists.

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        granter: Granter instance containing granted_entity_id, target_role_ids, and operations.
    """
    role_ids = granter.target_role_ids
    entity_id = granter.granted_entity_id

    if not role_ids:
        return

    # Insert object permissions (raises on conflict)
    await _insert_object_permissions(db_sess, role_ids, entity_id, granter.operations)



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.role import RoleData
from ai.backend.manager.models.rbac_models.role import RoleRow
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.services.permission_contoller.actions.base import RoleAction


@dataclass
class DeleteRoleAction(RoleAction):
    """Action for soft-deleting a role (status update)."""

    updater: Updater[RoleRow]

    @override
    def entity_id(self) -> str | None:
        return str(self.updater.pk_value)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DeleteRoleActionResult(BaseActionResult):
    data: RoleData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id)



"""Base types and utilities for repository layer.

Re-exports all public APIs for backward compatibility.
"""

from .creator import (
    BulkCreator,
    BulkCreatorError,
    BulkCreatorResult,
    BulkCreatorResultWithFailures,
    Creator,
    CreatorResult,
    CreatorSpec,
    execute_bulk_creator,
    execute_bulk_creator_partial,
    execute_creator,
)
from .export import (
    ExportDataStream,
    ExportFieldDef,
    ExportFieldType,
    ExportFormatter,
    ExportLimitExceeded,
    ExportQueryParams,
    ReportDef,
    StreamingExportQuery,
    execute_streaming_export,
)
from .pagination import (
    CursorBackwardPagination,
    CursorForwardPagination,
    NoPagination,
    OffsetPagination,
    PageInfoResult,
    QueryPagination,
)
from .purger import (
    BatchPurger,
    BatchPurgerResult,
    BatchPurgerSpec,
    Purger,
    PurgerResult,
    execute_batch_purger,
    execute_purger,
)
from .querier import (
    BatchQuerier,
    BatchQuerierResult,
    Querier,
    QuerierResult,
    execute_batch_querier,
    execute_querier,
)
from .types import (
    CursorConditionFactory,
    ExistenceCheck,
    QueryCondition,
    QueryOrder,
    SearchScope,
)
from .updater import (
    BatchUpdater,
    BatchUpdaterResult,
    BatchUpdaterSpec,
    Updater,
    UpdaterResult,
    UpdaterSpec,
    execute_batch_updater,
    execute_updater,
)
from .upserter import (
    BulkUpserter,
    BulkUpserterResult,
    Upserter,
    UpserterResult,
    UpserterSpec,
    execute_bulk_upserter,
    execute_upserter,
)
from .utils import (
    combine_conditions_or,
    negate_conditions,
)

__all__ = [
    # Types
    "QueryCondition",
    "QueryOrder",
    "CursorConditionFactory",
    "ExistenceCheck",
    "SearchScope",
    # Export
    "ExportDataStream",
    "ExportFieldDef",
    "ExportFieldType",
    "ExportFormatter",
    "ExportLimitExceeded",
    "ExportQueryParams",
    "ReportDef",
    "StreamingExportQuery",
    "execute_streaming_export",
    # Pagination
    "QueryPagination",
    "NoPagination",
    "OffsetPagination",
    "CursorForwardPagination",
    "CursorBackwardPagination",
    "PageInfoResult",
    # Querier
    "Querier",
    "QuerierResult",
    "execute_querier",
    # BatchQuerier
    "BatchQuerier",
    "BatchQuerierResult",
    "execute_batch_querier",
    # Creator
    "CreatorSpec",
    "Creator",
    "CreatorResult",
    "execute_creator",
    # BulkCreator
    "BulkCreator",
    "BulkCreatorError",
    "BulkCreatorResult",
    "BulkCreatorResultWithFailures",
    "execute_bulk_creator",
    "execute_bulk_creator_partial",
    # Updater
    "UpdaterSpec",
    "Updater",
    "UpdaterResult",
    "execute_updater",
    # BatchUpdater
    "BatchUpdaterSpec",
    "BatchUpdater",
    "BatchUpdaterResult",
    "execute_batch_updater",
    # Upserter
    "UpserterSpec",
    "Upserter",
    "UpserterResult",
    "execute_upserter",
    # BulkUpserter
    "BulkUpserter",
    "BulkUpserterResult",
    "execute_bulk_upserter",
    # Purger
    "Purger",
    "PurgerResult",
    "execute_purger",
    # BatchPurger
    "BatchPurgerSpec",
    "BatchPurger",
    "BatchPurgerResult",
    "execute_batch_purger",
    # Utils
    "combine_conditions_or",
    "negate_conditions",
]



"""Domain V2 GraphQL data fetcher functions."""

from __future__ import annotations

from functools import lru_cache

from strawberry import Info
from strawberry.relay import PageInfo

from ai.backend.manager.api.gql.adapter import PaginationOptions, PaginationSpec
from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.api.gql.domain_v2.types import (
    DomainV2Connection,
    DomainV2Edge,
    DomainV2Filter,
    DomainV2GQL,
    DomainV2OrderBy,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.models.domain.row import DomainRow
from ai.backend.manager.repositories.domain.options import DomainConditions, DomainOrders
from ai.backend.manager.repositories.domain.types import DomainSearchScope
from ai.backend.manager.services.domain.actions.get_domain import GetDomainAction
from ai.backend.manager.services.domain.actions.search_domains import SearchDomainsAction
from ai.backend.manager.services.domain.actions.search_rg_domains import SearchRGDomainsAction


@lru_cache(maxsize=1)
def get_domain_pagination_spec() -> PaginationSpec:
    """Cached pagination spec for domain queries."""
    return PaginationSpec(
        forward_order=DomainOrders.created_at(ascending=False),
        backward_order=DomainOrders.created_at(ascending=True),
        forward_condition_factory=DomainConditions.by_cursor_forward,
        backward_condition_factory=DomainConditions.by_cursor_backward,
        tiebreaker_order=DomainRow.name.asc(),
    )


async def fetch_domain(
    info: Info[StrawberryGQLContext],
    domain_name: str,
) -> DomainV2GQL:
    """Fetch a single domain by name.

    Args:
        info: Strawberry GraphQL context.
        domain_name: Name of the domain to retrieve.

    Returns:
        DomainV2GQL object.

    Raises:
        DomainNotFound: If the domain does not exist.
    """
    processors = info.context.processors

    # Execute via processor
    action_result = await processors.domain.get_domain.wait_for_complete(
        GetDomainAction(domain_name=domain_name)
    )

    return DomainV2GQL.from_data(action_result.data)


async def fetch_admin_domains(
    info: Info[StrawberryGQLContext],
    filter: DomainV2Filter | None = None,
    order_by: list[DomainV2OrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> DomainV2Connection:
    """Fetch all domains with optional filtering, ordering, and pagination.

    This is the admin-level fetcher that returns all domains in the system.

    Args:
        info: Strawberry GraphQL context.
        filter: Optional filter criteria.
        order_by: Optional ordering specification.
        before: Cursor for backward pagination.
        after: Cursor for forward pagination.
        first: Number of items from the start.
        last: Number of items from the end.
        limit: Maximum number of items (offset-based).
        offset: Starting position (offset-based).

    Returns:
        DomainV2Connection with paginated domain records.
    """
    processors = info.context.processors

    # Build querier
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        get_domain_pagination_spec(),
        filter=filter,
        order_by=order_by,
        base_conditions=None,
    )

    # Execute via processor
    action_result = await processors.domain.search_domains.wait_for_complete(
        SearchDomainsAction(querier=querier)
    )

    # Build connection
    nodes = [DomainV2GQL.from_data(data) for data in action_result.items]
    edges = [DomainV2Edge(node=node, cursor=encode_cursor(str(node.id))) for node in nodes]

    return DomainV2Connection(
        edges=edges,
        page_info=PageInfo(
            has_next_page=action_result.has_next_page,
            has_previous_page=action_result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=action_result.total_count,
    )


async def fetch_rg_domains(
    info: Info[StrawberryGQLContext],
    scope: DomainSearchScope,
    filter: DomainV2Filter | None = None,
    order_by: list[DomainV2OrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> DomainV2Connection:
    """Fetch domains within a resource group scope.

    This fetcher returns only domains that are associated with the specified resource group
    through the sgroups_for_domains relationship.

    Args:
        info: Strawberry GraphQL context.
        scope: DomainSearchScope containing resource_group filter.
        filter: Optional filter criteria.
        order_by: Optional ordering specification.
        before: Cursor for backward pagination.
        after: Cursor for forward pagination.
        first: Number of items from the start.
        last: Number of items from the end.
        limit: Maximum number of items (offset-based).
        offset: Starting position (offset-based).

    Returns:
        DomainV2Connection with paginated domain records.
    """
    processors = info.context.processors

    # Build querier
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        get_domain_pagination_spec(),
        filter=filter,
        order_by=order_by,
        base_conditions=None,
    )

    # Execute via processor with scope
    action_result = await processors.domain.search_rg_domains.wait_for_complete(
        SearchRGDomainsAction(scope=scope, querier=querier)
    )

    # Build connection
    nodes = [DomainV2GQL.from_data(data) for data in action_result.items]
    edges = [DomainV2Edge(node=node, cursor=encode_cursor(str(node.id))) for node in nodes]

    return DomainV2Connection(
        edges=edges,
        page_info=PageInfo(
            has_next_page=action_result.has_next_page,
            has_previous_page=action_result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=action_result.total_count,
    )



"""Domain-level app configuration actions."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.app_config.types import AppConfigData
from ai.backend.manager.repositories.app_config.updaters import AppConfigUpdaterSpec

from .base import AppConfigAction


@dataclass
class GetDomainConfigAction(AppConfigAction):
    """Action to get domain-level app configuration."""

    domain_name: str

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.APP_CONFIG_DOMAIN

    @override
    def entity_id(self) -> str | None:
        return self.domain_name

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class GetDomainConfigActionResult(BaseActionResult):
    """Result of get domain config action."""

    result: AppConfigData | None

    @override
    def entity_id(self) -> str | None:
        return self.result.scope_id if self.result else None


@dataclass
class UpsertDomainConfigAction(AppConfigAction):
    """Action to create or update domain-level app configuration."""

    domain_name: str
    updater_spec: AppConfigUpdaterSpec

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.APP_CONFIG_DOMAIN

    @override
    def entity_id(self) -> str | None:
        return self.domain_name

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class UpsertDomainConfigActionResult(BaseActionResult):
    """Result of upsert domain config action."""

    result: AppConfigData

    @override
    def entity_id(self) -> str | None:
        return self.result.scope_id


@dataclass
class DeleteDomainConfigAction(AppConfigAction):
    """Action to delete domain-level app configuration."""

    domain_name: str

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.APP_CONFIG_DOMAIN

    @override
    def entity_id(self) -> str | None:
        return self.domain_name

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DeleteDomainConfigActionResult(BaseActionResult):
    """Result of delete domain config action."""

    deleted: bool
    domain_name: str

    @override
    def entity_id(self) -> str | None:
        return self.domain_name if self.deleted else None



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.services.session.actions.commit_base import SessionCommitAction


@dataclass
class CommitSessionAction(SessionCommitAction):
    session_name: str
    owner_access_key: AccessKey
    filename: str | None

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CommitSessionActionResult(BaseActionResult):
    session_data: SessionData

    # TODO: Add proper type
    commit_result: Mapping[str, Any]

    @override
    def entity_id(self) -> str | None:
        return str(self.session_data.id)



from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import NotificationAction


@dataclass
class DeleteRuleAction(NotificationAction):
    """Action to delete a notification rule."""

    rule_id: UUID

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE

    @override
    def entity_id(self) -> str | None:
        return str(self.rule_id)


@dataclass
class DeleteRuleActionResult(BaseActionResult):
    """Result of deleting a notification rule."""

    deleted: bool

    @override
    def entity_id(self) -> str | None:
        return None



"""
Data classes for entity query operations in the RBAC system.
"""

from __future__ import annotations

from dataclasses import dataclass

from ai.backend.manager.data.common.types import SearchResult

from .types import EntityType


@dataclass(frozen=True)
class EntityData:
    """Information about an entity within a scope."""

    entity_type: EntityType
    entity_id: str


@dataclass(frozen=True)
class EntityListResult(SearchResult[EntityData]):
    """Result of entity search with pagination info."""

    pass



"""add_device_metadata_to_resource_slot_types

Revision ID: ccf8ae5c90fe
Revises: f41bbe0c0f12
Create Date: 2026-02-12 12:00:00.000000

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql as pgsql

# revision identifiers, used by Alembic.
revision = "ccf8ae5c90fe"
down_revision = "f41bbe0c0f12"
branch_labels = None
depends_on = None

# Slot types that did NOT exist in the original seed (6 original: cpu, mem,
# cuda.device, cuda.shares, rocm.device, tpu.device).
_new_slot_names = (
    "ipu.device",
    "atom.device",
    "atom-plus.device",
    "atom-max.device",
    "gaudi2.device",
    "warboy.device",
    "rngd.device",
    "hyperaccel-lpu.device",
)


def upgrade() -> None:
    # 1. Add new columns as nullable first (existing rows need defaults)
    op.add_column(
        "resource_slot_types",
        sa.Column("description", sa.String(length=256), nullable=True),
    )
    op.add_column(
        "resource_slot_types",
        sa.Column("display_unit", sa.String(length=32), nullable=True),
    )
    op.add_column(
        "resource_slot_types",
        sa.Column("display_icon", sa.String(length=64), nullable=True),
    )
    op.add_column(
        "resource_slot_types",
        sa.Column("number_format", pgsql.JSONB(), nullable=True),
    )

    # 2. Upsert all 14 slot types (6 existing + 8 new) with full metadata
    # Use exec_driver_sql to avoid sa.text() parsing JSON colons as bind params
    conn = op.get_bind()
    conn.exec_driver_sql("""
        INSERT INTO resource_slot_types
            (slot_name, slot_type, display_name, description,
             display_unit, display_icon, number_format, rank)
        VALUES
            ('cpu',        'count', 'CPU',        'CPU',
             'Core',    'cpu',     '{"binary":false,"round_length":0}', 100),
            ('mem',        'bytes', 'RAM',        'Memory',
             'GiB',     'ram',     '{"binary":true,"round_length":0}',  200),
            ('cuda.device','count', 'GPU',        'CUDA-capable GPU',
             'GPU',     'nvidia',  '{"binary":false,"round_length":0}', 300),
            ('cuda.shares','count', 'fGPU',       'CUDA-capable GPU (fractional)',
             'fGPU',    'nvidia',  '{"binary":false,"round_length":2}', 400),
            ('rocm.device','count', 'GPU',        'ROCm-capable GPU',
             'GPU',     'rocm',    '{"binary":false,"round_length":0}', 500),
            ('tpu.device', 'count', 'TPU',        'TPU device',
             'GPU',     'tpu',     '{"binary":false,"round_length":0}', 600),
            ('ipu.device', 'count', 'IPU',        'IPU device',
             'IPU',     'ipu',     '{"binary":false,"round_length":0}', 700),
            ('atom.device','count', 'ATOM Device','ATOM',
             'ATOM',    'rebel',   '{"binary":false,"round_length":0}', 800),
            ('atom-plus.device','count','ATOM+ Device','ATOM+',
             'ATOM+',   'rebel',   '{"binary":false,"round_length":0}', 900),
            ('atom-max.device','count','ATOM Max Device','ATOM Max',
             'ATOM Max','rebel',   '{"binary":false,"round_length":0}', 1000),
            ('gaudi2.device','count','Gaudi 2 Device','Gaudi 2',
             'Gaudi 2', 'gaudi',   '{"binary":false,"round_length":0}', 1100),
            ('warboy.device','count','Warboy Device','Furiosa Warboy',
             'Warboy',  'furiosa', '{"binary":false,"round_length":0}', 1200),
            ('rngd.device','count', 'RNGD Device','Furiosa RNGD',
             'RNGD',    'furiosa', '{"binary":false,"round_length":0}', 1300),
            ('hyperaccel-lpu.device','count','Hyperaccel LPU Device','Hyperaccel LPU',
             'LPU',     'lpu',     '{"binary":false,"round_length":0}', 1400)
        ON CONFLICT (slot_name) DO UPDATE SET
            display_name  = EXCLUDED.display_name,
            description   = EXCLUDED.description,
            display_unit  = EXCLUDED.display_unit,
            display_icon  = EXCLUDED.display_icon,
            number_format = EXCLUDED.number_format,
            rank          = EXCLUDED.rank
    """)

    # 3. Fill defaults for any dynamically-registered rows not covered by upsert
    conn.exec_driver_sql("""
        UPDATE resource_slot_types
        SET description  = COALESCE(description, ''),
            display_unit = COALESCE(display_unit, ''),
            display_icon = COALESCE(display_icon, ''),
            number_format = COALESCE(number_format, '{"binary":false,"round_length":0}'::jsonb)
        WHERE description IS NULL
           OR display_unit IS NULL
           OR display_icon IS NULL
           OR number_format IS NULL
    """)

    # 4. Also make display_name NOT NULL (was nullable before)
    conn.exec_driver_sql("""
        UPDATE resource_slot_types
        SET display_name = COALESCE(display_name, '')
        WHERE display_name IS NULL
    """)

    # 5. Set columns to NOT NULL with server defaults
    op.alter_column(
        "resource_slot_types",
        "display_name",
        nullable=False,
        server_default=sa.text("''"),
    )
    op.alter_column(
        "resource_slot_types",
        "description",
        nullable=False,
        server_default=sa.text("''"),
    )
    op.alter_column(
        "resource_slot_types",
        "display_unit",
        nullable=False,
        server_default=sa.text("''"),
    )
    op.alter_column(
        "resource_slot_types",
        "display_icon",
        nullable=False,
        server_default=sa.text("''"),
    )
    op.alter_column(
        "resource_slot_types",
        "number_format",
        nullable=False,
        server_default=sa.text(r"""'{"binary"\:false,"round_length"\:0}'::jsonb"""),
    )


def downgrade() -> None:
    conn = op.get_bind()

    # 1. Delete newly added slot types (keep original 6)
    conn.execute(
        sa.text("DELETE FROM resource_slot_types WHERE slot_name = ANY(:names)"),
        {"names": list(_new_slot_names)},
    )

    # 2. Revert original 6 display_name and rank values; restore display_name nullable
    conn.exec_driver_sql("""
        UPDATE resource_slot_types SET display_name = v.display_name, rank = v.rank
        FROM (VALUES
            ('cpu',         'CPU',        40),
            ('mem',         'Memory',     50),
            ('cuda.device', 'GPU (CUDA)', 10),
            ('cuda.shares', 'GPU (fGPU)', 20),
            ('rocm.device', 'GPU (ROCm)', 30),
            ('tpu.device',  'TPU',        35)
        ) AS v(slot_name, display_name, rank)
        WHERE resource_slot_types.slot_name = v.slot_name
    """)
    op.alter_column("resource_slot_types", "display_name", nullable=True, server_default=None)

    # 3. Drop new columns
    op.drop_column("resource_slot_types", "number_format")
    op.drop_column("resource_slot_types", "display_icon")
    op.drop_column("resource_slot_types", "display_unit")
    op.drop_column("resource_slot_types", "description")



"""Session-specific recorder context for scheduler coordinator."""

from __future__ import annotations

from ai.backend.common.types import SessionId
from ai.backend.manager.sokovan.recorder.context import RecorderContext

# Session-specific type alias
SessionRecorderContext = RecorderContext[SessionId]

__all__ = [
    "SessionRecorderContext",
]



from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseAction


class DeploymentReplicaBaseAction(BaseAction):
    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.DEPLOYMENT_REPLICA



import logging
import uuid
from collections.abc import Awaitable, Callable, Sequence
from datetime import UTC, datetime

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.actions.action import (
    BaseActionTriggerMeta,
)
from ai.backend.manager.actions.action.single_entity import (
    BaseSingleEntityAction,
    BaseSingleEntityActionResult,
)
from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.validator.single_entity import SingleEntityActionValidator

from .base import ActionRunner

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class SingleEntityActionProcessor[
    TSingleEntityAction: BaseSingleEntityAction,
    TSingleEntityActionResult: BaseSingleEntityActionResult,
]:
    _validators: Sequence[SingleEntityActionValidator]

    _runner: ActionRunner[TSingleEntityAction, TSingleEntityActionResult]

    def __init__(
        self,
        func: Callable[[TSingleEntityAction], Awaitable[TSingleEntityActionResult]],
        monitors: Sequence[ActionMonitor] | None = None,
        validators: Sequence[SingleEntityActionValidator] | None = None,
    ) -> None:
        self._runner = ActionRunner(func, monitors)

        self._validators = validators or []

    async def _run(self, action: TSingleEntityAction) -> TSingleEntityActionResult:
        started_at = datetime.now(UTC)
        action_id = uuid.uuid4()
        action_trigger_meta = BaseActionTriggerMeta(action_id=action_id, started_at=started_at)
        for validator in self._validators:
            await validator.validate(action, action_trigger_meta)

        return await self._runner.run(action, action_trigger_meta)

    async def wait_for_complete(self, action: TSingleEntityAction) -> TSingleEntityActionResult:
        return await self._run(action)



from .row import (
    ScalingGroupForDomainRow,
    ScalingGroupForKeypairsRow,
    ScalingGroupForProjectRow,
    ScalingGroupOpts,
    ScalingGroupPermissionContext,
    ScalingGroupPermissionContextBuilder,
    ScalingGroupRow,
    and_names,
    get_scaling_groups,
    query_allowed_sgroups,
    scaling_groups,
    sgroups_for_domains,
    sgroups_for_groups,
    sgroups_for_keypairs,
)
from .types import FairShareScalingGroupSpec

__all__ = (
    "FairShareScalingGroupSpec",
    "ScalingGroupForDomainRow",
    "ScalingGroupForKeypairsRow",
    "ScalingGroupForProjectRow",
    "ScalingGroupOpts",
    "ScalingGroupPermissionContext",
    "ScalingGroupPermissionContextBuilder",
    "ScalingGroupRow",
    "and_names",
    "get_scaling_groups",
    "query_allowed_sgroups",
    "scaling_groups",
    "sgroups_for_domains",
    "sgroups_for_groups",
    "sgroups_for_keypairs",
)



"""UpdaterSpec implementations for artifact repository."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.manager.models.artifact import ArtifactRow
from ai.backend.manager.repositories.base.updater import UpdaterSpec
from ai.backend.manager.types import TriState


@dataclass
class ArtifactUpdaterSpec(UpdaterSpec[ArtifactRow]):
    """UpdaterSpec for artifact updates."""

    readonly: TriState[bool] = field(default_factory=TriState[bool].nop)
    description: TriState[str] = field(default_factory=TriState[str].nop)

    @property
    @override
    def row_class(self) -> type[ArtifactRow]:
        return ArtifactRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.readonly.update_dict(to_update, "readonly")
        self.description.update_dict(to_update, "description")
        return to_update



"""UpdaterSpec implementations for session repository."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.repositories.base.updater import UpdaterSpec
from ai.backend.manager.types import OptionalState


@dataclass
class SessionUpdaterSpec(UpdaterSpec[SessionRow]):
    """UpdaterSpec for session updates."""

    name: OptionalState[str] = field(default_factory=OptionalState.nop)
    priority: OptionalState[int] = field(default_factory=OptionalState.nop)

    @property
    @override
    def row_class(self) -> type[SessionRow]:
        return SessionRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.priority.update_dict(to_update, "priority")
        return to_update



"""CreatorSpec implementations for app config entities."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, override

from ai.backend.manager.data.app_config.types import AppConfigScopeType
from ai.backend.manager.models.app_config import AppConfigRow
from ai.backend.manager.repositories.base.creator import CreatorSpec


@dataclass
class AppConfigCreatorSpec(CreatorSpec[AppConfigRow]):
    """CreatorSpec for app configurations."""

    scope_type: AppConfigScopeType
    scope_id: str
    extra_config: dict[str, Any]

    @override
    def build_row(self) -> AppConfigRow:
        return AppConfigRow(
            scope_type=self.scope_type,
            scope_id=self.scope_id,
            extra_config=self.extra_config,
        )



"""CreatorSpec implementations for scheduling history entities."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import override
from uuid import UUID

from ai.backend.common.data.endpoint.types import EndpointLifecycle
from ai.backend.common.types import KernelId, SessionId
from ai.backend.manager.data.deployment.types import RouteStatus
from ai.backend.manager.data.kernel.types import KernelSchedulingPhase
from ai.backend.manager.data.session.types import (
    SchedulingResult,
    SessionStatus,
    SubStepResult,
)
from ai.backend.manager.models.scheduling_history import (
    DeploymentHistoryRow,
    KernelSchedulingHistoryRow,
    RouteHistoryRow,
    SessionSchedulingHistoryRow,
)
from ai.backend.manager.repositories.base import CreatorSpec

__all__ = (
    "DeploymentHistoryCreatorSpec",
    "KernelSchedulingHistoryCreatorSpec",
    "RouteHistoryCreatorSpec",
    "SessionSchedulingHistoryCreatorSpec",
)


@dataclass
class SessionSchedulingHistoryCreatorSpec(CreatorSpec[SessionSchedulingHistoryRow]):
    """CreatorSpec for session scheduling history."""

    session_id: SessionId
    phase: str  # ScheduleType value
    result: SchedulingResult
    message: str
    from_status: SessionStatus | None = None
    to_status: SessionStatus | None = None
    error_code: str | None = None
    sub_steps: list[SubStepResult] = field(default_factory=list)

    @override
    def build_row(self) -> SessionSchedulingHistoryRow:
        return SessionSchedulingHistoryRow(
            session_id=self.session_id,
            phase=self.phase,
            from_status=str(self.from_status) if self.from_status else None,
            to_status=str(self.to_status) if self.to_status else None,
            result=str(self.result),
            error_code=self.error_code,
            message=self.message,
            sub_steps=self.sub_steps,  # PydanticListColumn handles serialization
            attempts=1,
        )


@dataclass
class KernelSchedulingHistoryCreatorSpec(CreatorSpec[KernelSchedulingHistoryRow]):
    """CreatorSpec for kernel scheduling history."""

    kernel_id: KernelId
    session_id: SessionId
    phase: str  # ScheduleType value
    result: SchedulingResult
    message: str
    from_status: KernelSchedulingPhase | None = None
    to_status: KernelSchedulingPhase | None = None
    error_code: str | None = None

    @override
    def build_row(self) -> KernelSchedulingHistoryRow:
        return KernelSchedulingHistoryRow(
            kernel_id=self.kernel_id,
            session_id=self.session_id,
            phase=self.phase,
            from_status=str(self.from_status) if self.from_status else None,
            to_status=str(self.to_status) if self.to_status else None,
            result=str(self.result),
            error_code=self.error_code,
            message=self.message,
            attempts=1,
        )


@dataclass
class DeploymentHistoryCreatorSpec(CreatorSpec[DeploymentHistoryRow]):
    """CreatorSpec for deployment history."""

    deployment_id: UUID
    phase: str  # DeploymentLifecycleType value
    result: SchedulingResult
    message: str
    from_status: EndpointLifecycle | None = None
    to_status: EndpointLifecycle | None = None
    error_code: str | None = None
    sub_steps: list[SubStepResult] = field(default_factory=list)

    @override
    def build_row(self) -> DeploymentHistoryRow:
        return DeploymentHistoryRow(
            deployment_id=self.deployment_id,
            phase=self.phase,
            from_status=str(self.from_status.value) if self.from_status else None,
            to_status=str(self.to_status.value) if self.to_status else None,
            result=str(self.result),
            error_code=self.error_code,
            message=self.message,
            sub_steps=self.sub_steps,
            attempts=1,
        )


@dataclass
class RouteHistoryCreatorSpec(CreatorSpec[RouteHistoryRow]):
    """CreatorSpec for route history."""

    route_id: UUID
    deployment_id: UUID
    phase: str  # RouteLifecycleType value
    result: SchedulingResult
    message: str
    from_status: RouteStatus | None = None
    to_status: RouteStatus | None = None
    error_code: str | None = None
    sub_steps: list[SubStepResult] = field(default_factory=list)

    @override
    def build_row(self) -> RouteHistoryRow:
        return RouteHistoryRow(
            route_id=self.route_id,
            deployment_id=self.deployment_id,
            phase=self.phase,
            from_status=str(self.from_status.value) if self.from_status else None,
            to_status=str(self.to_status.value) if self.to_status else None,
            result=str(self.result),
            error_code=self.error_code,
            message=self.message,
            sub_steps=self.sub_steps,
            attempts=1,
        )



"""ProjectV2 GraphQL nested types for structured field groups."""

from __future__ import annotations

from datetime import datetime

import strawberry

from .enums import ProjectTypeEnum, VFolderHostPermissionEnum

# ============================================================================
# Basic Information
# ============================================================================


@strawberry.type(
    name="ProjectBasicInfo",
    description=(
        "Added in 26.2.0. Basic project information. "
        "Contains identity and descriptive fields for the project."
    ),
)
class ProjectBasicInfoGQL:
    """Basic project information."""

    name: str = strawberry.field(description="Project name.")
    description: str | None = strawberry.field(description="Optional description of the project.")
    type: ProjectTypeEnum = strawberry.field(
        description="Project type determining its purpose. See ProjectTypeV2 enum."
    )
    integration_id: str | None = strawberry.field(
        description="External system integration identifier."
    )


# ============================================================================
# Organization Context
# ============================================================================


@strawberry.type(
    name="ProjectOrganizationInfo",
    description=(
        "Added in 26.2.0. Project's organizational context. "
        "Contains domain membership and resource policy information."
    ),
)
class ProjectOrganizationInfoGQL:
    """Project's organizational context."""

    domain_name: str = strawberry.field(description="Name of the domain this project belongs to.")
    resource_policy: str = strawberry.field(
        description="Name of the project resource policy applied to this project."
    )


# ============================================================================
# Storage Configuration
# ============================================================================


@strawberry.type(
    name="VFolderHostPermissionEntry",
    description=(
        "Added in 26.2.0. Storage host permission configuration. "
        "Defines what operations are allowed for a specific storage host."
    ),
)
class VFolderHostPermissionEntryGQL:
    """Storage host permission entry."""

    host: str = strawberry.field(
        description="Storage host identifier (e.g., 'default', 'storage-01')."
    )
    permissions: list[VFolderHostPermissionEnum] = strawberry.field(
        description=(
            "List of permissions granted for this host. "
            "See VFolderHostPermissionV2 enum for available permissions."
        )
    )


@strawberry.type(
    name="ProjectStorageInfo",
    description=(
        "Added in 26.2.0. Project storage configuration. "
        "Contains allowed virtual folder hosts and their permissions."
    ),
)
class ProjectStorageInfoGQL:
    """Project storage configuration."""

    allowed_vfolder_hosts: list[VFolderHostPermissionEntryGQL] = strawberry.field(
        description=(
            "Storage hosts accessible to this project with their permissions. "
            "Each entry specifies a host and the operations allowed on it. "
            "Empty list means no storage access."
        )
    )


# ============================================================================
# Lifecycle Information
# ============================================================================


@strawberry.type(
    name="ProjectLifecycleInfo",
    description=(
        "Added in 26.2.0. Project lifecycle information. "
        "Contains activation status and timestamp tracking."
    ),
)
class ProjectLifecycleInfoGQL:
    """Project lifecycle information."""

    is_active: bool | None = strawberry.field(
        description=(
            "Whether the project is active. "
            "Inactive projects cannot create new sessions or perform operations."
        )
    )
    created_at: datetime | None = strawberry.field(
        description="Timestamp when the project was created."
    )
    modified_at: datetime | None = strawberry.field(
        description="Timestamp when the project was last modified."
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.types import BinarySize, ResourceSlot
from ai.backend.manager.models.resource_preset import ResourcePresetRow
from ai.backend.manager.repositories.base.creator import CreatorSpec


@dataclass
class ResourcePresetCreatorSpec(CreatorSpec[ResourcePresetRow]):
    """CreatorSpec for resource preset."""

    name: str
    resource_slots: ResourceSlot
    shared_memory: str | None
    scaling_group_name: str | None

    @override
    def build_row(self) -> ResourcePresetRow:
        row = ResourcePresetRow()
        row.name = self.name
        row.resource_slots = self.resource_slots
        row.shared_memory = (
            int(BinarySize.from_str(self.shared_memory)) if self.shared_memory else None
        )
        row.scaling_group_name = self.scaling_group_name
        return row



from __future__ import annotations

import copy
import logging
import uuid
from collections.abc import Sequence
from datetime import datetime
from decimal import Decimal
from typing import Any, cast
from uuid import UUID

import aiotools
import msgpack
import sqlalchemy as sa
from sqlalchemy.engine import CursorResult
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.exception import InvalidAPIParameters
from ai.backend.common.types import SlotName, VFolderID
from ai.backend.common.utils import nmget
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.group.types import GroupData
from ai.backend.manager.errors.resource import (
    ProjectHasActiveEndpointsError,
    ProjectHasActiveKernelsError,
    ProjectHasVFoldersMountedError,
    ProjectNotFound,
)
from ai.backend.manager.models.domain import domains
from ai.backend.manager.models.endpoint import EndpointLifecycle, EndpointRow
from ai.backend.manager.models.group import association_groups_users, groups
from ai.backend.manager.models.group.row import (
    AssocGroupUserRow,
    GroupRow,
)
from ai.backend.manager.models.kernel import (
    AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
    LIVE_STATUS,
    RESOURCE_USAGE_KERNEL_STATUSES,
    KernelRow,
    kernels,
)
from ai.backend.manager.models.resource_policy import project_resource_policies
from ai.backend.manager.models.resource_usage import fetch_resource_usage
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.user import users
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.models.vfolder import (
    VFolderDeletionInfo,
    VFolderRow,
    VFolderStatusSet,
    initiate_vfolder_deletion,
    vfolder_status_map,
    vfolders,
)
from ai.backend.manager.repositories.base.creator import Creator, execute_creator
from ai.backend.manager.repositories.base.purger import BatchPurger, execute_batch_purger
from ai.backend.manager.repositories.base.querier import BatchQuerier, execute_batch_querier
from ai.backend.manager.repositories.base.updater import Updater, execute_updater
from ai.backend.manager.repositories.group.creators import GroupCreatorSpec
from ai.backend.manager.repositories.group.purgers import (
    GroupBatchPurgerSpec,
    GroupEndpointBatchPurgerSpec,
    GroupKernelBatchPurgerSpec,
    GroupSessionBatchPurgerSpec,
    SessionByIdsBatchPurgerSpec,
)
from ai.backend.manager.repositories.group.types import (
    DomainProjectSearchScope,
    GroupSearchResult,
    UserProjectSearchScope,
)
from ai.backend.manager.repositories.permission_controller.role_manager import RoleManager

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class GroupDBSource:
    _db: ExtendedAsyncSAEngine
    _role_manager: RoleManager

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db
        self._role_manager = RoleManager()

    async def create(self, creator: Creator[GroupRow]) -> GroupData:
        """Create a new group."""
        spec = cast(GroupCreatorSpec, creator.spec)
        async with self._db.begin_session() as db_session:
            # Validate domain exists
            domain_exists = await db_session.scalar(
                sa.select(sa.exists().where(domains.c.name == spec.domain_name))
            )
            if not domain_exists:
                raise InvalidAPIParameters(
                    f"Cannot create group: Domain '{spec.domain_name}' does not exist"
                )

            # Validate resource policy exists
            policy_exists = await db_session.scalar(
                sa.select(
                    sa.exists().where(project_resource_policies.c.name == spec.resource_policy)
                )
            )
            if not policy_exists:
                raise InvalidAPIParameters(
                    f"Cannot create group: Resource policy '{spec.resource_policy}' does not exist"
                )

            # Check if group already exists
            check_stmt = sa.select(GroupRow).where(
                sa.and_(
                    GroupRow.name == spec.name,
                    GroupRow.domain_name == spec.domain_name,
                )
            )
            existing_group = await db_session.scalar(check_stmt)
            if existing_group is not None:
                raise InvalidAPIParameters(
                    f"Group with name '{spec.name}' already exists in domain '{spec.domain_name}'"
                )

            # Create the group
            creator_result = await execute_creator(db_session, creator)
            row: GroupRow = creator_result.row
            data = row.to_data()
            # Create RBAC role and permissions for the group
            await self._role_manager.create_system_role(db_session, data)

            return data

    async def modify_validated(
        self,
        updater: Updater[GroupRow],
        user_update_mode: str | None = None,
        user_uuids: list[uuid.UUID] | None = None,
    ) -> GroupData | None:
        """Modify a group with validation."""
        group_id = updater.pk_value

        async with self._db.begin_session() as session:
            # First verify the group exists
            existing_group = await session.scalar(
                sa.select(groups.c.id).where(groups.c.id == group_id)
            )
            if existing_group is None:
                raise ProjectNotFound(f"Group not found: {group_id}")

            # Handle user addition/removal
            if user_uuids and user_update_mode:
                if user_update_mode == "add":
                    values = [{"user_id": uuid, "group_id": group_id} for uuid in user_uuids]
                    await session.execute(
                        sa.insert(association_groups_users).values(values),
                    )
                elif user_update_mode == "remove":
                    await session.execute(
                        sa.delete(association_groups_users).where(
                            (association_groups_users.c.user_id.in_(user_uuids))
                            & (association_groups_users.c.group_id == group_id),
                        ),
                    )

            # Update group data (execute_updater returns None if no values to update)
            result = await execute_updater(session, updater)
            if result is not None:
                return result.row.to_data()

            # No group updates or only user updates were performed
            return None

    async def mark_inactive(self, group_id: uuid.UUID) -> None:
        """Mark a group as inactive (soft delete)."""
        async with self._db.begin_session() as session:
            result = await session.execute(
                sa.update(groups)
                .values(
                    is_active=False,
                    integration_id=None,
                )
                .where(groups.c.id == group_id)
            )
            if cast(CursorResult[Any], result).rowcount > 0:
                return
            raise ProjectNotFound(f"Group not found: {group_id}")

    async def get_container_stats_for_period(
        self,
        start_date: datetime,
        end_date: datetime,
        valkey_stat_client: ValkeyStatClient,
        config_provider: ManagerConfigProvider,
        group_ids: Sequence[UUID] | None = None,
    ) -> list[dict[str, Any]]:
        """Get container statistics for groups within a time period."""
        async with self._db.begin_readonly() as conn:
            j = kernels.join(groups, groups.c.id == kernels.c.group_id).join(
                users, users.c.uuid == kernels.c.user_uuid
            )
            query = (
                sa.select(
                    kernels.c.id,
                    kernels.c.container_id,
                    kernels.c.session_id,
                    kernels.c.session_name,
                    kernels.c.access_key,
                    kernels.c.agent,
                    kernels.c.domain_name,
                    kernels.c.group_id,
                    kernels.c.attached_devices,
                    kernels.c.occupied_slots,
                    kernels.c.resource_opts,
                    kernels.c.vfolder_mounts,
                    kernels.c.mounts,
                    kernels.c.image,
                    kernels.c.status,
                    kernels.c.status_info,
                    kernels.c.status_changed,
                    kernels.c.last_stat,
                    kernels.c.status_history,
                    kernels.c.created_at,
                    kernels.c.terminated_at,
                    kernels.c.cluster_mode,
                    groups.c.name,
                    users.c.email,
                    users.c.full_name,
                )
                .select_from(j)
                .where(
                    # Filter sessions which existence period overlaps with requested period
                    (
                        (kernels.c.terminated_at >= start_date)
                        & (kernels.c.created_at < end_date)
                        & (kernels.c.status.in_(RESOURCE_USAGE_KERNEL_STATUSES))
                    )
                    |
                    # Or, filter running sessions which created before requested end_date
                    ((kernels.c.created_at < end_date) & (kernels.c.status.in_(LIVE_STATUS))),
                )
                .order_by(sa.asc(kernels.c.terminated_at))
            )
            if group_ids:
                query = query.where(kernels.c.group_id.in_(group_ids))
            result = await conn.execute(query)
            rows = result.fetchall()

        kernel_ids = [str(row.id) for row in rows]
        raw_stats = await valkey_stat_client.get_user_kernel_statistics_batch(kernel_ids)

        objs_per_group = {}
        local_tz = config_provider.config.system.timezone

        for row, raw_stat in zip(rows, raw_stats, strict=True):
            group_id = str(row.group_id)
            last_stat = row.last_stat
            if not last_stat:
                if raw_stat is None:
                    log.warning("stat object for {} not found on redis, skipping", str(row.id))
                    continue
                last_stat = msgpack.unpackb(raw_stat)
            nfs = None
            if row.vfolder_mounts:
                # For >=22.03, return used host directories instead of volume host, which is not so useful.
                nfs = list({str(mount.host_path) for mount in row.vfolder_mounts})
            elif row.mounts and isinstance(row.mounts[0], list):
                # For the kernel records that have legacy contents of `mounts`.
                nfs = list({mount[2] for mount in row.mounts})
            if row.terminated_at is None:
                used_time = used_days = None
            else:
                used_time = str(row.terminated_at - row.created_at)
                used_days = (
                    row.terminated_at.astimezone(local_tz).toordinal()
                    - row.created_at.astimezone(local_tz).toordinal()
                    + 1
                )
            device_type = set()
            gpu_smp_allocated = 0
            gpu_mem_allocated = 0
            if row.attached_devices and row.attached_devices.get("cuda"):
                for dev_info in row.attached_devices["cuda"]:
                    if dev_info.get("model_name"):
                        device_type.add(dev_info["model_name"])
                    gpu_smp_allocated += int(nmget(dev_info, "data.smp", 0))
                    gpu_mem_allocated += int(nmget(dev_info, "data.mem", 0))
            gpu_allocated = Decimal(0)
            for key, value in row.occupied_slots.items():
                if SlotName(key).is_accelerator():
                    gpu_allocated += value
            c_info = {
                "id": str(row.id),
                "session_id": str(row.session_id),
                "container_id": row.container_id,
                "domain_name": row.domain_name,
                "group_id": str(row.group_id),
                "group_name": row.name,
                "name": row.session_name,
                "access_key": row.access_key,
                "email": row.email,
                "full_name": row.full_name,
                "agent": row.agent,
                "cpu_allocated": float(row.occupied_slots.get("cpu", 0)),
                "cpu_used": float(nmget(last_stat, "cpu_used.current", 0)),
                "mem_allocated": int(row.occupied_slots.get("mem", 0)),
                "mem_used": int(nmget(last_stat, "mem.capacity", 0)),
                "shared_memory": int(nmget(row.resource_opts, "shmem", 0)),
                "disk_allocated": 0,  # TODO: disk quota limit
                "disk_used": int(nmget(last_stat, "io_scratch_size/stats.max", 0, "/")),
                "io_read": int(nmget(last_stat, "io_read.current", 0)),
                "io_write": int(nmget(last_stat, "io_write.current", 0)),
                "used_time": used_time,
                "used_days": used_days,
                "device_type": list(device_type),
                "smp": float(gpu_smp_allocated),
                "gpu_mem_allocated": float(gpu_mem_allocated),
                "gpu_allocated": float(gpu_allocated),
                "nfs": nfs,
                "image_id": row.image,  # TODO: image id
                "image_name": row.image,
                "created_at": str(row.created_at),
                "terminated_at": str(row.terminated_at),
                "status": row.status.name,
                "status_info": row.status_info,
                "status_changed": str(row.status_changed),
                "status_history": row.status_history or {},
                "cluster_mode": row.cluster_mode,
            }
            if group_id not in objs_per_group:
                objs_per_group[group_id] = {
                    "domain_name": row.domain_name,
                    "g_id": group_id,
                    "g_name": row.name,  # this is group's name
                    "g_cpu_allocated": c_info["cpu_allocated"],
                    "g_cpu_used": c_info["cpu_used"],
                    "g_mem_allocated": c_info["mem_allocated"],
                    "g_mem_used": c_info["mem_used"],
                    "g_shared_memory": c_info["shared_memory"],
                    "g_disk_allocated": c_info["disk_allocated"],
                    "g_disk_used": c_info["disk_used"],
                    "g_io_read": c_info["io_read"],
                    "g_io_write": c_info["io_write"],
                    "g_device_type": copy.deepcopy(c_info["device_type"]),
                    "g_smp": c_info["smp"],
                    "g_gpu_mem_allocated": c_info["gpu_mem_allocated"],
                    "g_gpu_allocated": c_info["gpu_allocated"],
                    "c_infos": [c_info],
                }
            else:
                objs_per_group[group_id]["g_cpu_allocated"] += c_info["cpu_allocated"]
                objs_per_group[group_id]["g_cpu_used"] += c_info["cpu_used"]
                objs_per_group[group_id]["g_mem_allocated"] += c_info["mem_allocated"]
                objs_per_group[group_id]["g_mem_used"] += c_info["mem_used"]
                objs_per_group[group_id]["g_shared_memory"] += c_info["shared_memory"]
                objs_per_group[group_id]["g_disk_allocated"] += c_info["disk_allocated"]
                objs_per_group[group_id]["g_disk_used"] += c_info["disk_used"]
                objs_per_group[group_id]["g_io_read"] += c_info["io_read"]
                objs_per_group[group_id]["g_io_write"] += c_info["io_write"]
                for device in c_info["device_type"]:
                    if device not in objs_per_group[group_id]["g_device_type"]:
                        g_dev_type = objs_per_group[group_id]["g_device_type"]
                        g_dev_type.append(device)
                        objs_per_group[group_id]["g_device_type"] = list(set(g_dev_type))
                objs_per_group[group_id]["g_smp"] += c_info["smp"]
                objs_per_group[group_id]["g_gpu_mem_allocated"] += c_info["gpu_mem_allocated"]
                objs_per_group[group_id]["g_gpu_allocated"] += c_info["gpu_allocated"]
                objs_per_group[group_id]["c_infos"].append(c_info)
        return list(objs_per_group.values())

    async def fetch_project_resource_usage(
        self,
        start_date: datetime,
        end_date: datetime,
        project_ids: Sequence[UUID] | None = None,
    ) -> list[KernelRow]:
        """Fetch resource usage data for projects."""
        return await fetch_resource_usage(self._db, start_date, end_date, project_ids=project_ids)

    async def purge_group(
        self,
        group_id: uuid.UUID,
        storage_manager: StorageSessionManager,
    ) -> bool:
        """Completely remove a group and all its associated data."""
        async with self._db.begin_session() as session:
            # Pre-flight checks
            if await self._check_group_vfolders_mounted_to_active_kernels(session, group_id):
                raise ProjectHasVFoldersMountedError(
                    f"error on deleting project {group_id} with vfolders mounted to active kernels"
                )

            if await self._check_group_has_active_kernels(session, group_id):
                raise ProjectHasActiveKernelsError(
                    f"error on deleting project {group_id} with active kernels"
                )

            # Delete associated resources
            await self._delete_group_endpoints(session, group_id)

            # Commit session before vfolder deletion (which uses separate transactions)
            await session.commit()

        # Delete vfolders (uses separate transaction)
        await self._delete_group_vfolders(group_id, storage_manager)

        async with self._db.begin_session() as session:
            # Delete remaining data
            await self._delete_group_kernels(session, group_id)
            await self._delete_group_sessions(session, group_id)

            # Finally delete the group itself
            result = await execute_batch_purger(
                session, BatchPurger(spec=GroupBatchPurgerSpec(group_id=group_id), batch_size=1)
            )

            if result.deleted_count > 0:
                return True
            raise ProjectNotFound("project not found")

    async def _check_group_vfolders_mounted_to_active_kernels(
        self, session: SASession, group_id: uuid.UUID
    ) -> bool:
        """Check if group has vfolders mounted to active kernels."""
        # Get group vfolder IDs
        query = sa.select(vfolders.c.id).select_from(vfolders).where(vfolders.c.group == group_id)
        result = await session.execute(query)
        rows = result.fetchall()
        group_vfolder_ids = [row.id for row in rows]

        # Check if any active kernels have these vfolders mounted
        query = (
            sa.select(kernels.c.mounts)
            .select_from(kernels)
            .where(
                (kernels.c.group_id == group_id)
                & (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES)),
            )
        )
        async for row in await session.stream(query):
            for _mount in row.mounts:
                try:
                    vfolder_id = uuid.UUID(_mount[2])
                    if vfolder_id in group_vfolder_ids:
                        return True
                except Exception:
                    log.warning("Malformed mount entry in group {}, skipping: {}", group_id, _mount)
        return False

    async def _check_group_has_active_kernels(
        self, session: SASession, group_id: uuid.UUID
    ) -> bool:
        """Check if group has active kernels."""
        query = (
            sa.select(sa.func.count())
            .select_from(kernels)
            .where(
                (kernels.c.group_id == group_id)
                & (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES))
            )
        )
        active_kernel_count = await session.scalar(query)
        return (active_kernel_count or 0) > 0

    async def _delete_group_vfolders(
        self,
        group_id: uuid.UUID,
        storage_manager: StorageSessionManager,
    ) -> int:
        """Delete all vfolders belonging to the group."""
        target_vfs: list[VFolderDeletionInfo] = []
        async with self._db.begin_session() as session:
            query = sa.select(VFolderRow).where(
                sa.and_(
                    VFolderRow.group == group_id,
                    VFolderRow.status.in_(vfolder_status_map[VFolderStatusSet.DELETABLE]),
                )
            )
            result = await session.scalars(query)
            rows = cast(list[VFolderRow], result.fetchall())
            for vf in rows:
                target_vfs.append(
                    VFolderDeletionInfo(VFolderID.from_row(vf), vf.host, vf.unmanaged_path)
                )

        storage_ptask_group = aiotools.PersistentTaskGroup()
        await initiate_vfolder_deletion(
            self._db,
            target_vfs,
            storage_manager,
            storage_ptask_group,
        )

        return len(target_vfs)

    async def _delete_group_kernels(self, session: SASession, group_id: uuid.UUID) -> int:
        """Delete all kernels belonging to the group."""
        result = await execute_batch_purger(
            session, BatchPurger(spec=GroupKernelBatchPurgerSpec(group_id=group_id))
        )
        return result.deleted_count

    async def _delete_group_sessions(self, session: SASession, group_id: uuid.UUID) -> int:
        """Delete all sessions belonging to the group."""
        result = await execute_batch_purger(
            session, BatchPurger(spec=GroupSessionBatchPurgerSpec(group_id=group_id))
        )
        return result.deleted_count

    async def _delete_group_endpoints(self, session: SASession, group_id: uuid.UUID) -> None:
        """Delete all endpoints belonging to the group."""
        # Get all endpoints for the group to check for active ones
        endpoints = (
            await session.execute(
                sa.select(
                    EndpointRow.id,
                    sa.case(
                        (
                            EndpointRow.lifecycle_stage.in_([
                                EndpointLifecycle.CREATED,
                                EndpointLifecycle.DESTROYING,
                            ]),
                            True,
                        ),
                        else_=False,
                    ).label("is_active"),
                ).where(EndpointRow.project == group_id)
            )
        ).all()

        if len(endpoints) == 0:
            return

        # Check for active endpoints
        active_endpoints = [ep.id for ep in endpoints if ep.is_active]
        if len(active_endpoints) > 0:
            raise ProjectHasActiveEndpointsError(f"project {group_id} has active endpoints")

        # Collect session IDs first (must be done before endpoint/routing deletion
        # as the query depends on RoutingRow)
        endpoint_ids = [ep.id for ep in endpoints]
        session_id_query = sa.select(RoutingRow.session).where(
            sa.and_(
                RoutingRow.endpoint.in_(endpoint_ids),
                RoutingRow.session.is_not(None),
            )
        )
        session_ids_result = await session.scalars(session_id_query)
        session_ids = [sid for sid in session_ids_result.all() if sid is not None]

        # Delete endpoints first (routings are CASCADE deleted automatically)
        await execute_batch_purger(
            session, BatchPurger(spec=GroupEndpointBatchPurgerSpec(project_id=group_id))
        )

        # Delete sessions using the collected IDs
        if session_ids:
            await execute_batch_purger(
                session, BatchPurger(spec=SessionByIdsBatchPurgerSpec(session_ids=session_ids))
            )

    async def get_project(self, project_id: UUID) -> GroupData:
        """Get a single project by UUID.

        Args:
            project_id: UUID of the project.

        Returns:
            GroupData for the project.

        Raises:
            ProjectNotFound: If project does not exist.
        """
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(sa.select(GroupRow).where(GroupRow.id == project_id))
            row = result.scalar_one_or_none()
            if row is None:
                raise ProjectNotFound(f"Project {project_id} not found")
            return row.to_data()

    async def search_projects(
        self,
        querier: BatchQuerier,
    ) -> GroupSearchResult:
        """Search all projects (admin only).

        Args:
            querier: Contains conditions, orders, and pagination.

        Returns:
            GroupSearchResult with items, total_count, and pagination flags.
        """
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(GroupRow)
            result = await execute_batch_querier(db_sess, query, querier)

            items = [row.GroupRow.to_data() for row in result.rows]

            return GroupSearchResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_projects_by_domain(
        self,
        scope: DomainProjectSearchScope,
        querier: BatchQuerier,
    ) -> GroupSearchResult:
        """Search projects within a domain.

        Args:
            scope: DomainProjectSearchScope defining the domain to search within.
            querier: Contains conditions, orders, and pagination.

        Returns:
            GroupSearchResult with items, total_count, and pagination flags.
        """
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(GroupRow)
            result = await execute_batch_querier(db_sess, query, querier, scope=scope)

            items = [row.GroupRow.to_data() for row in result.rows]

            return GroupSearchResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_projects_by_user(
        self,
        scope: UserProjectSearchScope,
        querier: BatchQuerier,
    ) -> GroupSearchResult:
        """Search projects a user is member of.

        Joins with association_groups_users to find user's projects.

        Args:
            scope: UserProjectSearchScope defining the user to search for.
            querier: Contains conditions, orders, and pagination.

        Returns:
            GroupSearchResult with items, total_count, and pagination flags.
        """
        async with self._db.begin_readonly_session() as db_sess:
            query = (
                sa.select(GroupRow)
                .select_from(GroupRow)
                .join(AssocGroupUserRow, GroupRow.id == AssocGroupUserRow.group_id)
            )
            result = await execute_batch_querier(db_sess, query, querier, scope=scope)

            items = [row.GroupRow.to_data() for row in result.rows]

            return GroupSearchResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )
